@@ -58,7 +58,6 @@ get_stringr_pattern_options <- function(pattern) {
   }
 
   ensure_opts <- function(opts) {
-
     # default options for the simple cases
     if (is.character(opts)) {
       opts <- list(pattern = opts, fixed = FALSE, ignore_case = FALSE)
@@ -352,6 +351,12 @@ register_bindings_string_regex <- function() {
   # Encapsulate some common logic for sub/gsub/str_replace/str_replace_all
   arrow_r_string_replace_function <- function(max_replacements) {
     function(pattern, replacement, x, ignore.case = FALSE, fixed = FALSE) {
+      if (length(pattern) != 1) {
+        stop("`pattern` must be a length 1 character vector")
+      }
+      if (length(replacement) != 1) {
+        stop("`replacement` must be a length 1 character vector")
+      }
       Expression$create(
         ifelse(fixed && !ignore.case, "replace_substring", "replace_substring_regex"),
         x,
@@ -516,7 +521,7 @@ register_bindings_string_other <- function() {
         msg = "`stop` must be length 1 - other lengths are not supported in Arrow"
       )
 
-      # substr treats values as if they're on a continous number line, so values
+      # substr treats values as if they're on a continuous number line, so values
       # 0 are effectively blank characters - set `start` to 1 here so Arrow mimics
       # this behavior
       if (start <= 0) {
